// __multiversion__
// This signals the loading code to prepend either #version 100 or #version 300 es as apropriate.

#include "fragmentVersionCentroid.h"

#ifdef GL_FRAGMENT_PRECISION_HIGH
    #define hmp highp
#else
    #define hmp mediump
#endif

#if __VERSION__ >= 300
	#ifndef BYPASS_PIXEL_SHADER
		#if defined(TEXEL_AA) && defined(TEXEL_AA_FEATURE)
			_centroid in highp vec2 uv0;
			_centroid in highp vec2 uv1;
		#else
			_centroid in vec2 uv0;
			_centroid in vec2 uv1;
		#endif
	#endif
#else
	#ifndef BYPASS_PIXEL_SHADER
		varying vec2 uv0;
		varying vec2 uv1;
	#endif
#endif

varying vec4 color;

#ifdef FOG
varying vec4 fogColor;
#endif

varying hmp vec3 cPos;
varying hmp vec3 wPos;
flat varying float water;

uniform hmp float TOTAL_REAL_WORLD_TIME;
uniform vec4 FOG_COLOR;
uniform vec2 FOG_CONTROL;

#include "uniformShaderConstants.h"
#include "util.h"
#include "simplexNoise.glsl"

LAYOUT_BINDING(0) uniform sampler2D TEXTURE_0;
LAYOUT_BINDING(1) uniform sampler2D TEXTURE_1;
LAYOUT_BINDING(2) uniform sampler2D TEXTURE_2;

struct Fancy_Misty_Colour {
	vec3 skylight;
	vec3 sunSetlight;
	vec3 sunDaylight;
	vec3 moonlight;
	vec3 torchlight;
	vec3 wetColour;
	vec3 waterDayColour;
	vec3 waterNightColour;
};

struct Fancy_Misty_Data {
	float  daylight;
	float  moonlight;
	float  duskOrDawn;
	float  outDoor;
	bool   underwater;
	float  rain;
	vec2   shadowlight;
	float  diffuseLight;
};

struct Fancy_Misty_Settings {
	float  skyBrightness;
	float  sunBrightness;
	float  sunSetBrightness;
	float  moonBrightness;
	float  torchBrightness;
	float  shadowBrightness;
	vec3   sunPosition;
};

// Copied from:
// https://knarkowicz.wordpress.com/2016/01/06/aces-filmic-tone-mapping-curve/
vec3 ACESFilm(vec3 x) {
    return clamp((x * (2.51 * x + 0.03)) / (x * (2.43 * x + 0.59) + 0.14), 0.0, 1.0);
}

// See:
// https://www.shadertoy.com/view/MsKXDh
hmp float waterWave(hmp vec2 coord, hmp float time) {
    hmp vec2 p = coord; 
    p.x += snoise(vec2(p.y * 1.5 + time * 2.0)) / 40.0;
    p.y += cos   (p.x * 1.0 + time * 2.0) / 45.0;   
    p.x += sin   ((p.y + p.x) * 1.5 + time * 2.0) / (9.00 + (2.0 * sin(time)));
    p.y += snoise(vec2((p.y + p.x) * 1.5 + time * 2.0)) / (10.0 + (2.0 * cos(time)));
            
    return smoothstep(0.05, 1.00, clamp(abs(snoise(p)), 0.0, 1.0));
}

hmp float getWaveHeight(hmp vec2 uv) {
  return waterWave(uv, TOTAL_REAL_WORLD_TIME * 0.5);
}

hmp vec4 waterWavenormal(hmp vec2 uv, vec2 resolution, float scale) {
	hmp vec2 step = 1.0 / resolution;
	hmp float height = getWaveHeight(uv);
	hmp vec2 dxy = height - vec2(getWaveHeight(uv + vec2(step.x, 0.0)),
  		getWaveHeight(uv + vec2(0.0, step.y)));
    
	return vec4(normalize(vec3(dxy * scale / step, 1.0)), height);
}

vec3 textureHeightNormal(vec2 uv, vec2 resolution, float scale) {
	vec2 step = 1.0 / resolution;
	float height = texture2D(TEXTURE_0, uv).r;
	vec2 dxy = height - vec2(texture2D(TEXTURE_0, uv + vec2(step.x, 0.0)).r,
  		texture2D(TEXTURE_0, uv + vec2(0.0, step.y)).r);
    
	return normalize(vec3(dxy * scale / step, 1.0));
}

float drawTheSun(vec3 inPos){
   return inversesqrt(inPos.x * inPos.x + inPos.y * inPos.y + inPos.z * inPos.z);   
}

#define DAYLIGHT texture2D(TEXTURE_1, vec2(0.0, 1.0)).r
#define OUTDOOR smoothstep(0.850, 0.875, uv1.y)
#ifdef FOG
	#define UNDERWATER step(FOG_CONTROL.x, 0.0)
	#define RAIN smoothstep(0.690, 0.300, FOG_CONTROL.x)
#else
	#define UNDERWATER 0.0
	#define RAIN 0.0
#endif
#define WORLD_NORMAL normalize(cross(dFdx(wPos), dFdy(wPos)))

void main() {
#ifdef BYPASS_PIXEL_SHADER
	gl_FragColor = vec4(0.0);
	return;
#else

Fancy_Misty_Settings settings = Fancy_Misty_Settings(
    1.0,                 // Sky brightness
    0.0,                 // Sun brightness
    0.8,                 // Sunset brightness
    0.7,                 // Moon brightness
    0.0,                 // Torch brightness
    0.7,                 // Shadow brightness
    vec3(0.5, 1.0, -0.5) // Sun position
);

Fancy_Misty_Colour colour = Fancy_Misty_Colour(
    vec3(0.8392, 0.9098, 0.9961), // Skylight
    vec3(1.0000, 0.3569, 0.0196), // Sunsetlight
    vec3(1.0000, 0.8706, 0.7039), // Sundaylight
    vec3(0.44,   0.53,   1.0),    // Moonlight
    vec3(2.0,    1.26,   0.68),   // Torchlight
    vec3(0.7,    0.8,    0.9),    // Wet Colour
    vec3(0.0,    0.4,    0.5),    // Water colour day
    vec3(0.00,   0.05,   0.10)    // Water colour night
);

Fancy_Misty_Data data = Fancy_Misty_Data(
    mix(0.0, smoothstep(0.4, 1.0, DAYLIGHT), OUTDOOR),         // Daylight
    mix(0.0, 1.0 - DAYLIGHT, OUTDOOR),                         // Moonlight
    mix(0.0, min(smoothstep(0.495, 0.580, DAYLIGHT), 
        smoothstep(0.800, 0.600, DAYLIGHT)), OUTDOOR),         // Dusk or dawn
    OUTDOOR,                                                   // Outdoor
    bool(UNDERWATER),                                          // Underwater
    mix(0.0, bool(UNDERWATER) ? 1.0 : RAIN, OUTDOOR),          // Rain
    vec2(min(1.0, pow(uv1.x, 2.7)),
        mix(0.0, mix(OUTDOOR, 1.0, 1.0 - DAYLIGHT), uv1.y)),   // Shadowlight
    mix(1.0, dot(WORLD_NORMAL, settings.sunPosition), OUTDOOR) // Diffuse light
);

#if USE_TEXEL_AA
	vec4 diffuse = texture2D_AA(TEXTURE_0, uv0);
#else
	vec4 diffuse = texture2D(TEXTURE_0, uv0);
#endif
	
#ifdef SEASONS_FAR
	diffuse.a = 1.0;
#endif

#if USE_ALPHA_TEST
	#ifdef ALPHA_TO_COVERAGE
	#define ALPHA_THRESHOLD 0.05
	#else
	#define ALPHA_THRESHOLD 0.5
	#endif
	if(diffuse.a < ALPHA_THRESHOLD)
		discard;
#endif
	
vec4 inColor = color;

#if defined(BLEND)
	diffuse.a *= inColor.a;
#endif

#if !defined(ALWAYS_LIT)
	diffuse *= texture2D( TEXTURE_1, vec2(data.shadowlight.x, uv1.y) );
#endif

#ifndef SEASONS
	#if !USE_ALPHA_TEST && !defined(BLEND)
		diffuse.a = inColor.a;
	#endif
	
	diffuse.rgb *= inColor.rgb;
#else
	vec2 uv = inColor.xy;
	diffuse.rgb *= mix(vec3(1.0,1.0,1.0), texture2D( TEXTURE_2, uv).rgb*2.0, inColor.b);
	diffuse.rgb *= inColor.aaa;
	diffuse.a = 1.0;
#endif

if (water > 0.5 || bool(UNDERWATER)) {
	hmp vec3 refPos = cPos * 5.0;
	refPos /= cPos.y * 0.8;
	
	if (!bool(UNDERWATER)) {
		hmp float cosTheta = abs(dot(vec3(0.0, 1.0, 0.0), normalize(-cPos)));
		hmp float lenAms = clamp(length(normalize(cPos).xz * 0.08 / normalize(cPos).y), 0.0, 1.0);
		hmp float sun = mix(mix(0.0, clamp(drawTheSun(cross(refPos + dot(settings.sunPosition, waterWavenormal(wPos.xz * 0.5, vec2(10.0), 0.02).rgb) * 80.0 * 
		refPos / length(refPos) * 0.1, settings.sunPosition) * 3.0), 0.0, 1.0), smoothstep(0.4, 1.0, DAYLIGHT)), 0.0, data.rain);
		hmp float lightRef = max(mix(mix(max(0.0, fBM((refPos.xz + dot(settings.sunPosition, waterWavenormal(wPos.xz * 0.5, vec2(10.0), 0.02).rgb) * 
			25.0 * refPos.xz / length(refPos.xz)) * 0.1, 0.2, 1.0, 2, TOTAL_REAL_WORLD_TIME)), 0.0, cosTheta), 0.0, lenAms) * 0.65,
			sun * sun * sun);
		hmp vec4 refLight = vec4(texture2D(TEXTURE_1, uv1) + FOG_COLOR) * 0.5;
		
		diffuse.rgb  = mix(colour.waterNightColour, colour.waterDayColour, smoothstep(0.4, 1.0, DAYLIGHT) * uv1.y);
		diffuse      = mix(diffuse, mix(mix(diffuse, refLight, lenAms * 0.5), 
			refLight, lightRef), OUTDOOR);
	} else {
		diffuse.rgb = mix(diffuse.rgb, diffuse.rgb * 1.5, 
			1.0 - abs(waterWavenormal(wPos.xz * 0.4, vec2(10.0), 0.02).a));
	}
}

diffuse.rgb *= mix(vec3(1.0), settings.skyBrightness + colour.skylight, mix(data.daylight, 0.0, data.rain));         // Skylight
diffuse.rgb *= mix(vec3(1.0), settings.sunBrightness + colour.sunDaylight, mix(data.daylight, 0.0, data.rain));      // Sunlight
diffuse.rgb *= mix(vec3(1.0), settings.sunSetBrightness + colour.sunSetlight, mix(data.duskOrDawn, 0.0, data.rain)); // Sunsetlight
diffuse.rgb *= mix(vec3(1.0), settings.moonBrightness + colour.moonlight, mix(data.moonlight, 0.0, data.rain));      // Moonlight
diffuse.rgb *= mix(vec3(1.0), settings.torchBrightness + colour.torchlight, data.shadowlight.x                       // Torchlight
	* (1.0 - smoothstep(0.6, 0.8, uv1.y) * smoothstep(0.4, 1.0, DAYLIGHT)));
diffuse.rgb *= mix(settings.shadowBrightness, 1.0,                                                                     // Shadow
	mix(data.shadowlight.y, 1.0, data.shadowlight.x));
diffuse.rgb *= mix(0.95, 1.0,
	dot(settings.sunPosition, textureHeightNormal(uv0, vec2(4092.0), 1.0)));                                           // Texture height lighting
diffuse.rgb *= mix(settings.shadowBrightness, 1.0, mix(data.diffuseLight, 1.0, data.shadowlight.x));                   // Diffuse lighting

diffuse.rgb = ACESFilm(diffuse.rgb); // Tone map

#ifdef FOG
	diffuse.rgb = mix( diffuse.rgb, fogColor.rgb, fogColor.a );
#endif

	gl_FragColor = diffuse;
	
#endif // BYPASS_PIXEL_SHADER
}
