#include "ShaderConstants.fxh"
#include "util.fxh"
#include "simplexNoise.hlsl"

struct PS_Input {
	float4 position : SV_Position;
#ifndef BYPASS_PIXEL_SHADER
	lpfloat4 color : COLOR;
	snorm float2 uv0 : TEXCOORD_0_FB_MSAA;
	snorm float2 uv1 : TEXCOORD_1_FB_MSAA;
#endif
	float3 wPos : worldPos;
	float3 cPos : cameraPos;
	float water : waterFlag;
#ifdef FOG
	float4 fogColor : FOG_COLOR;
#endif
};

struct PS_Output {
	float4 color : SV_Target;
};

struct Fancy_Misty_Colour {
	float3 skylight;
	float3 sunSetlight;
	float3 sunDaylight;
	float3 moonlight;
	float3 torchlight;
	float3 wetColour;
	float3 waterDayColour;
	float3 waterNightColour;
};

struct Fancy_Misty_Data {
	float  daylight;
	float  moonlight;
	float  duskOrDawn;
	float  outDoor;
	bool   underwater;
	float  rain;
	float2 shadowlight;
	float  diffuseLight;
};

struct Fancy_Misty_Settings {
	float  skyBrightness;
	float  sunBrightness;
	float  sunSetBrightness;
	float  moonBrightness;
	float  torchBrightness;
	float  shadowBrightness;
	float3 sunPosition;
};

// Copied from:
// https://knarkowicz.wordpress.com/2016/01/06/aces-filmic-tone-mapping-curve/
float3 ACESFilm(float3 x) {
    return saturate((x * (2.51f * x + 0.03f)) / (x * (2.43f * x + 0.59f) + 0.14f));
}

// See:
// https://www.shadertoy.com/view/MsKXDh
float waterWave(const float2 coord, const float time) {
    float2 p = coord; 
    p.x += snoise(p.y * 1.5 + time * 2.0) / 40.0;
    p.y += cos   (p.x * 1.0 + time * 2.0) / 45.0;   
    p.x += sin   ((p.y + p.x) * 1.5 + time * 2.0) / (9.00 + (2.0 * sin(time)));
    p.y += snoise((p.y + p.x) * 1.5 + time * 2.0) / (10.0 + (2.0 * cos(time)));
            
    return smoothstep(0.05, 1.00, clamp(abs(snoise(p)), 0.0, 1.0));
}

float getWaveHeight(float2 uv) {
  return waterWave(uv, TOTAL_REAL_WORLD_TIME * 0.5);
}

float4 waterWavenormal(float2 uv, float2 resolution, float scale) {
	float2 step = 1.0 / resolution;
	float height = getWaveHeight(uv);
	float2 dxy = height - float2(getWaveHeight(uv + float2(step.x, 0.0)),
  		getWaveHeight(uv + float2(0.0, step.y)));
    
	return float4(normalize(float3(dxy * scale / step, 1.0)), height);
}

float3 textureHeightNormal(float2 uv, float2 resolution, float scale) {
	float2 step = 1.0 / resolution;
	float height = TEXTURE_0.Sample(TextureSampler0, uv).r;
	float2 dxy = height - float2(TEXTURE_0.Sample(TextureSampler0, uv + float2(step.x, 0.0)).r,
  		TEXTURE_0.Sample(TextureSampler0, uv + float2(0.0, step.y)).r);
    
	return normalize(float3(dxy * scale / step, 1.0));
}

float drawTheSun(float3 inPos){
   return rsqrt(inPos.x * inPos.x + inPos.y * inPos.y + inPos.z * inPos.z);   
}

#define DAYLIGHT TEXTURE_1.Sample(TextureSampler1, float2(0.0, 1.0)).r
#define OUTDOOR smoothstep(0.850, 0.875, PSInput.uv1.y)
#ifdef FOG
	#define UNDERWATER step(FOG_CONTROL.x, 0.0)
	#define RAIN smoothstep(0.690, 0.300, FOG_CONTROL.x)
#else
	#define UNDERWATER 0.0
	#define RAIN 0.0
#endif
#define WORLD_NORMAL normalize(cross(ddx(-PSInput.wPos), ddy(PSInput.wPos)))

ROOT_SIGNATURE
void main(in PS_Input PSInput, out PS_Output PSOutput) {
#ifdef BYPASS_PIXEL_SHADER
    PSOutput.color = float4(0.0f, 0.0f, 0.0f, 0.0f);
    return;
#else

struct Fancy_Misty_Settings settings = {
	{
		1.0,                   // Sky brightness
		0.0,                   // Sun brightness
		0.8,                   // Sunset brightness
		0.7,                   // Moon brightness
		0.0,                   // Torch brightness
		0.7,                   // Shadow brightness
		float3(0.5, 1.0, -0.5) // Sun position
	}
};

struct Fancy_Misty_Colour colour = {
	{	
		float3(0.8392, 0.9098, 0.9961), // Skylight
		float3(1.0000, 0.3569, 0.0196), // Sunsetlight
		float3(1.0000, 0.8706, 0.7039), // Sundaylight
		float3(0.44,   0.53,   1.0),    // Moonlight
		float3(2.0,    1.26,   0.68),   // Torchlight
		float3(0.7,    0.8,    0.9),    // Wet Colour
		float3(0.0,    0.4,    0.5),    // Water colour day
		float3(0.00,    0.05,   0.10)   // Water colour night
	}
};

struct Fancy_Misty_Data data = {
	{
		lerp(0.0, smoothstep(0.4, 1.0, DAYLIGHT), OUTDOOR),                  // Daylight
		lerp(0.0, 1.0 - DAYLIGHT, OUTDOOR),                                  // Moonlight
		lerp(0.0, min(smoothstep(0.495, 0.580, DAYLIGHT), 
			smoothstep(0.800, 0.600, DAYLIGHT)), OUTDOOR),                   // Dusk or dawn
		OUTDOOR,                                                             // Outdoor
		bool(UNDERWATER),                                                    // Underwater
		lerp(0.0, bool(UNDERWATER) ? 1.0 : RAIN, OUTDOOR),                   // Rain
		float2(min(1.0, pow(PSInput.uv1.x, 2.7)),
			lerp(0.0, lerp(OUTDOOR, 1.0, 1.0 - DAYLIGHT), PSInput.uv1.y)),   // Shadowlight
		lerp(1.0, dot(WORLD_NORMAL, settings.sunPosition), OUTDOOR)          // Diffuse light
	}
};

#if USE_TEXEL_AA
	float4 diffuse = texture2D_AA(TEXTURE_0, TextureSampler0, PSInput.uv0 );
#else
	float4 diffuse = TEXTURE_0.Sample(TextureSampler0, PSInput.uv0);
#endif

#ifdef SEASONS_FAR
	diffuse.a = 1.0f;
#endif

#if USE_ALPHA_TEST
	#ifdef ALPHA_TO_COVERAGE
		#define ALPHA_THRESHOLD 0.05
	#else
		#define ALPHA_THRESHOLD 0.5
	#endif
	if(diffuse.a < ALPHA_THRESHOLD)
		discard;
#endif

#if defined(BLEND)
	diffuse.a *= PSInput.color.a;
#endif

#if !defined(ALWAYS_LIT)
	diffuse = diffuse * TEXTURE_1.Sample(TextureSampler1, float2(data.shadowlight.x, PSInput.uv1.y));
#endif

#ifndef SEASONS
	#if !USE_ALPHA_TEST && !defined(BLEND)
		diffuse.a = PSInput.color.a;
	#endif	

	diffuse.rgb *= PSInput.color.rgb;
#else
	float2 uv = PSInput.color.xy;
	diffuse.rgb *= lerp(1.0f, TEXTURE_2.Sample(TextureSampler2, uv).rgb*2.0f, PSInput.color.b);
	diffuse.rgb *= PSInput.color.aaa;
	diffuse.a = 1.0f;
#endif

if (PSInput.water > 0.5 || bool(UNDERWATER)) {
	float3 refPos = PSInput.cPos * 5.0;
	refPos /= PSInput.cPos.y * 0.8;
	
	if (!bool(UNDERWATER)) {
		float3 skp = (refPos + dot(settings.sunPosition, waterWavenormal(PSInput.wPos.xz * 0.5, 10.0, 0.02).rgb) * 
			25.0 * refPos / length(refPos)) * 0.1;
		float cosTheta = abs(dot(float3(0.0, 1.0, 0.0), normalize(-PSInput.cPos)));
		float lenAms = clamp(length(normalize(PSInput.cPos).xz * 0.08 / normalize(PSInput.cPos).y), 0.0, 1.0);
		float sun = lerp(lerp(0.0, clamp(drawTheSun(cross(refPos + dot(settings.sunPosition, waterWavenormal(PSInput.wPos.xz * 0.5, 10.0, 0.02).rgb) * 80.0 * 
		refPos / length(refPos) * 0.1, settings.sunPosition) * 3.0), 0.0, 1.0), smoothstep(0.4, 1.0, DAYLIGHT)), 0.0, data.rain);
		float lightRef = max(lerp(lerp(max(0.0, fBM(skp.xz, 0.2, 1.0, 2, TOTAL_REAL_WORLD_TIME)), 0.0, cosTheta), 0.0, lenAms) * 0.65,
			sun * sun * sun);
		float4 refLight = float4(TEXTURE_1.Sample(TextureSampler1, PSInput.uv1) + FOG_COLOR) * 0.5;
		
		diffuse.rgb  = lerp(colour.waterNightColour, colour.waterDayColour, smoothstep(0.4, 1.0, DAYLIGHT) * PSInput.uv1.y);
		diffuse      = lerp(diffuse, lerp(lerp(diffuse, refLight, lenAms * 0.5), 
			refLight, lightRef), OUTDOOR);
	} else {
		diffuse.rgb = lerp(diffuse.rgb, diffuse.rgb * 1.5, 
			1.0 - abs(waterWavenormal(PSInput.wPos.xz * 0.4, 10.0, 0.02).a));
	}
}

diffuse.rgb *= lerp(1.0, settings.skyBrightness + colour.skylight, lerp(data.daylight, 0.0, data.rain));         // Skylight
diffuse.rgb *= lerp(1.0, settings.sunBrightness + colour.sunDaylight, lerp(data.daylight, 0.0, data.rain));      // Sunlight
diffuse.rgb *= lerp(1.0, settings.sunSetBrightness + colour.sunSetlight, lerp(data.duskOrDawn, 0.0, data.rain)); // Sunsetlight
diffuse.rgb *= lerp(1.0, settings.moonBrightness + colour.moonlight, lerp(data.moonlight, 0.0, data.rain));      // Moonlight
diffuse.rgb *= lerp(1.0, settings.torchBrightness + colour.torchlight, data.shadowlight.x                        // Torchlight
	* (1.0 - smoothstep(0.6, 0.8, PSInput.uv1.y) * smoothstep(0.4, 1.0, DAYLIGHT)));
diffuse.rgb *= lerp(settings.shadowBrightness, 1.0,                                                              // Shadow
	lerp(data.shadowlight.y, 1.0, data.shadowlight.x));
diffuse.rgb *= lerp(0.95, 1.0,
	dot(settings.sunPosition, textureHeightNormal(PSInput.uv0, 4092.0, 1.0)));                                   // Texture height lighting
diffuse.rgb *= lerp(settings.shadowBrightness, 1.0, lerp(data.diffuseLight, 1.0, data.shadowlight.x));           // Diffuse lighting

diffuse.rgb = ACESFilm(diffuse.rgb); // Tone map

#ifdef FOG
	diffuse.rgb = lerp( diffuse.rgb, PSInput.fogColor.rgb, PSInput.fogColor.a );
#endif

	PSOutput.color = diffuse;

#ifdef VR_MODE
	PSOutput.color = max(PSOutput.color, 1 / 255.0f);
#endif

#endif // BYPASS_PIXEL_SHADER
}